# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any


def robocopy(src: Any, dest: Any): ...
def link_dir(link: Any, target: Any) -> None: ...
def link_dirs(link_target_tuples: Any) -> None: ...
def link_file(link: Any, target: Any) -> None: ...
def link_files(link_target_tuples: Any) -> None: ...
def unlink_dir(link: Any) -> None: ...
def unlink_dirs(links: Any) -> None: ...
def unlink_file(link: Any) -> None: ...
def unlink_files(links: Any) -> None: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Iterable
from typing import Union

Flint = Union[int, float]  # float or int
Paths = Iterable[str]  # iterable of path-strings



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from pupy._typing import Flint


def fmt_bytes(num: Flint) -> str: ...
def fmt_file_size(filepath: str) -> str: ...
def fmt_seconds(t1: float, t2: Optional[float] = ...) -> str: ...
def path2name(path_str: str) -> str: ...
def parent_dirpath(fdpath: str) -> str: ...
def timestamp(ts: Optional[float] = ...) -> str: ...
def ls(dirpath: str = ..., abs: bool = ...): ...
def ls_files(dirpath: Any, abs: bool = ...): ...
def ls_dirs(dirpath: str = ..., abs: bool = ...) -> Any: ...
def ls_files_dirs(dirpath: str) -> Tuple[List[str], List[str]]: ...
def link_dir(link: Any, target: Any): ...
def link_dirs(link_target_tuples: Any): ...
def link_file(link: Any, target: Any): ...
def link_files(link_target_tuples: Any): ...
def unlink_dir(link_path: str) -> Any: ...
def unlink_dirs(link_paths: Any): ...
def unlink_file(link: Any): ...
def unlink_files(links: Any): ...
def sync(src: Any, dest: Any): ...

class LinkedTmpDir:
    dirpath: Any = ...
    dirname: Any = ...
    file_links: Any = ...
    dir_links: Any = ...
    def __init__(
        self,
        suffix: Optional[Any] = ...,
        prefix: Optional[Any] = ...,
        dir: Optional[Any] = ...,
        link_targets: Optional[Any] = ...,
    ) -> None: ...
    def __enter__(self): ...
    def __exit__(self, exc: Any, value: Any, tb: Any) -> None: ...
    def cleanup(self) -> None: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from codecs import getwriter
from io import open
from itertools import count
from os import path
from os import utime
from typing import Dict
from typing import List

from pupy.decorations import mkdirs

try:
    from ujson import dump
    from ujson import load
except:
    from json import dump
    from json import load


def safepath(path_str: str) -> str:
    """Checks if a file/dir path is save/unused; returns an unused path.

    :param path_str: file or dir path
    :return: A file/dir path that does not exist and contains the given path

    """
    if path.exists(path_str):
        f_bn, f_ext = path.splitext(path_str)
        for n in count(1):
            safe_save_path = f_bn + "_({}).".format(str(n)) + f_ext
            if not path.exists(safe_save_path):
                return safe_save_path
    return path_str


def lbytes(filepath: str) -> None:
    """Read bytes from file path

    :param filepath: filepath as as string to read bites from
    :return: some bytes...
    """
    with open(filepath, "rb") as file:
        return file.read()


@mkdirs
def sstring(filepath: str, string: str) -> None:
    """Writes a string to filepath


    :param filepath: Filepath save location
    :param string: File as a string to be saved
    :return: None? what do you want? confirmation?

    .. note:: Decorated w/ @mkdirs
        Function is decorated with @mkdirs decorator; @mkdirs creates parent
        directories for the given filepath if they do not already exist.

    """
    with open(filepath, "wb") as file:
        file.write(string.encode("utf-8"))


def savings(filepath: str, string: str) -> None:
    """Alias for sstring"""
    return sstring(filepath, string)


def sstr(filepath: str, string: str) -> None:
    """Alias for sstring"""
    return sstring(filepath, string)


def lstring(filepath: str):
    """(lstring) Read and return the file-contents as a string given a filepath

    :param filepath: Path to a file to read
    :return: Content of the file read as a string

    """
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError:
        with open(filepath, "r", encoding="latin2") as f:
            return f.read()


def lstr(filepath):
    """Alias for lstring"""
    return lstring(filepath)


@mkdirs
def sjson(filepath: str, data: Dict[str, List[str]], min: bool = False) -> None:
    """Save json-serial-ize-able data to a specific filepath.

    :param filepath: destination filepath
    :param data: json cereal-izable dictionary/list/thing
    :param min: Bool flag -- minify the json file
    :return: None

    """
    if type(data) == dict and any(type(val) == bytes for val in data.values()):
        data = {k: str(v, encoding="utf-8") for k, v in data.items()}
    with open(filepath, "wb") as jsonfile:
        if min:
            dump(data, getwriter("utf-8")(jsonfile), ensure_ascii=False)
        else:
            dump(
                data,
                getwriter("utf-8")(jsonfile),
                indent=4,
                sort_keys=True,
                ensure_ascii=False,
            )


def save_jasm(filepath: str, data: Dict[str, List[str]], min: bool = False) -> None:
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)


def sjasm(filepath: str, data: Dict[str, List[str]], min: bool = False) -> None:
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)


def spak(filepath: str, data: Dict[str, List[str]], min: bool = False) -> None:
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)


@mkdirs
def ljson(filepath: str):
    """Load a json file given a filepath and return the file-data

    :param filepath: path to the jasm file you want to load
    :return: Loaded file contents
    """

    with open(filepath) as infile:
        return load(infile)


def load_jasm(filepath: str) -> Dict[str, List[str]]:
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)


def ljasm(filepath: str) -> Dict[str, List[str]]:
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)


def lpak(filepath: str) -> Dict[str, List[str]]:
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)


@mkdirs
def touch(filepath: str) -> None:
    """Touches a file just like touch on the command line

    :param filepath: filepath to 'touch' in a unix-y sense
    :return: None
    """
    with open(filepath, "a"):
        utime(filepath, None)


def shebang(filepath: str):
    """returns the shebang path given a filepath or None if it does not exist.

    :param filepath: path to a file w/ a shebange line
    :return: shebang line or None

    .. doctest::python

        >>> from inspect import getabsfile
        >>> from pupy.savings_n_loads import sstr
        >>> script = 'ashellscript.sh'
        >>> sstr(script, '#!/bin/bash\\necho "howdy"\\n')
        >>> shebang(script)
        '#!/bin/bash'
        >>> from os import remove
        >>> remove(script)

    """
    with open(filepath, "r") as f:
        first = f.readline().strip("\n")
        return first if first[:2] == "#!" else None



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import Optional


def unescaped_str(arg_str: Any): ...

parser: Any

def main(args: Optional[Any] = ...) -> None: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import Iterator
from typing import List


def yesno(question: Any, default: bool = ..., tries: int = ...): ...
def term_table(
    strings: List[str], row_wise: bool = ..., filler: str = ...
) -> Iterator[Any]: ...



# -*- coding: utf-8 -*-

from subprocess import PIPE
from subprocess import run


def robocopy(src, dest):
    """Robocopy for sheldon

    :param dest: path to local tdir
    :param src: path to remote (raid) tdir
    :return: subprocess return code from robocopy

    Robocopy return codes:

    0. No files were copied. No failure was encountered. No files were
       mismatched. The files already exist in the destination directory;
       therefore, the copy operation was skipped.
    1. All files were copied successfully.
    2. There are some additional files in the destination directory that are not
       present in the source directory. No files were copied.
    3. Some files were copied. Additional files were present. No failure
       was encountered.
    5. Some files were copied. Some files were mismatched. No failure was
       encountered.
    6. Additional files and mismatched files exist. No files were copied
       and no failures were encountered. This means that the files already
       exist in the destination directory.
    7. Files were copied, a file mismatch was present, and additional files
       were present.
    8. Several files did not copy.


    """
    subproc = run(["robocopy", src, dest, "/mir", "/mt"], stdout=PIPE, stderr=PIPE)
    if subproc.returncode not in (0, 1, 2, 3):
        print(subproc)
    return subproc.returncode


def link_dir(link, target):
    run(args=["mklink", "/D", link, target], stdout=PIPE, stderr=PIPE, shell=True)


def link_dirs(link_target_tuples):
    link_args = []
    for link_target_tuple in link_target_tuples:
        link_args.extend(["mklink", "/D", *link_target_tuple, "&&"])
    run(args=link_args[:-1], stdout=PIPE, stderr=PIPE, shell=True)


def link_file(link, target):
    run(args=["mklink", link, target], stdout=PIPE, stderr=PIPE, shell=True)


def link_files(link_target_tuples):
    link_args = []
    for link_target_tuple in link_target_tuples:
        link_args.extend(["mklink", *link_target_tuple, "&&"])
    run(args=link_args[:-1], stdout=PIPE, stderr=PIPE, shell=True)


def unlink_dir(link):
    run(args=["RD", link], stdout=PIPE, stderr=PIPE, shell=True)


def unlink_dirs(links):
    cmd_args = " && ".join("RD {}".format(link) for link in links).split(" ")
    run(args=cmd_args, stdout=PIPE, stderr=PIPE, shell=True)


def unlink_file(link):
    run(args=["Del", link], stdout=PIPE, stderr=PIPE, shell=True)


def unlink_files(links):
    cmd_args = " && ".join("Del {}".format(link) for link in links).split(" ")
    run(args=cmd_args, stdout=PIPE, stderr=PIPE, shell=True)



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

from pupy._typing import Flint


def files_gen(dirpath: str = ..., abs: bool = ...) -> Iterator[Any]: ...
def dirs_gen(dirpath: str = ..., abs: bool = ...) -> Iterator[Any]: ...
def exhaust(it: Iterable[Any]) -> None: ...
def chunks(it: Union[List[int], str], chunk_size: int) -> Iterator[Any]: ...
def is_permutation(
    a: Union[int, List[int], str], b: Union[int, List[int], str]
) -> bool: ...
def rotate(rlist: List[int], rn: int = ..., left_rotate: bool = ...) -> List[int]: ...
def rotations_gen(rlist: Tuple[int, int, int, int]) -> Iterator[Any]: ...
def digits_list(number: int) -> List[int]: ...
def int_from_digits(digits: Iterable[int]) -> int: ...
def iter_product(l: Iterable[int]) -> Flint: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from os import symlink
from os import unlink
from subprocess import PIPE
from subprocess import run


def rsync(src: str, dest: str, delete: bool = False):
    """Sheldon rsync wrapper for syncing tdirs

    :param delete:
    :param dest: path to local tdir
    :param src: path to remote (raid) tdir
    :return: subprocess return code from rsync

    Rsync return codes:

     - 0 == Success
     - 1 == Syntax or usage error
     - 2 == Protocol incompatibility
     - 3 == Errors selecting input/output files, dirs
     - 4 == Requested  action not supported: an attempt was made to manipulate
       64-bit files on a platform that cannot support them; or an option
       was specified that is supported by the client and not the server.
     - 5 == Error starting client-server protocol
     - 6 == Daemon unable to append to log-file
     - 10 == Error in socket I/O
     - 11 == Error in file I/O
     - 12 == Error in rsync protocol data stream
     - 13 == Errors with program diagnostics
     - 14 == Error in IPC code
     - 20 == Received SIGUSR1 or SIGINT
     - 21 == Some error returned by waitpid()
     - 22 == Error allocating core memory buffers
     - 23 == Partial transfer due to error
     - 24 == Partial transfer due to vanished source files
     - 25 == The --max-delete limit stopped deletions
     - 30 == Timeout in data send/receive
     - 35 == Timeout waiting for daemon connection

    """
    subproc = run(["mkdir", "-p", dest], stdout=PIPE, stderr=PIPE)
    if not dest.endswith("/"):
        dest = "{}/".format(dest)
    if not src.endswith("/"):
        src = "{}/".format(src)
    rsync_args = [
        "rsync",
        "-a",
        "-O",
        "--no-o",
        "--no-g",
        "--no-p",
        "--delete" if delete else None,
        src,
        dest,
    ]
    subproc = run(args=list(filter(None, rsync_args)), stdout=PIPE, stderr=PIPE)
    return subproc


def link_dir(linkpath, targetpath):
    symlink(targetpath, linkpath)


def link_dirs(link_target_tuples):
    for link, target in link_target_tuples:
        link_dir(link, target)


def link_file(linkpath: str, targetpath: str) -> None:
    symlink(targetpath, linkpath)


def link_files(link_target_tuples):
    for link, target in link_target_tuples:
        link_file(link, target)


def unlink_dir(link):
    unlink(link)


def unlink_dirs(links):
    for link in links:
        unlink(link)


def unlink_file(link):
    unlink(link)


def unlink_files(links):
    for link in links:
        unlink(link)



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import Dict
from typing import List


def safepath(path_str: str) -> str: ...
def lbytes(filepath: str) -> None: ...
def sstring(filepath: str, string: str) -> None: ...
def savings(filepath: str, string: str) -> None: ...
def sstr(filepath: str, string: str) -> None: ...
def lstring(filepath: str) -> Any: ...
def lstr(filepath: Any): ...
def sjson(filepath: str, data: Dict[str, List[str]], min: bool = ...) -> None: ...
def save_jasm(filepath: str, data: Dict[str, List[str]], min: bool = ...) -> None: ...
def sjasm(filepath: str, data: Dict[str, List[str]], min: bool = ...) -> None: ...
def spak(filepath: str, data: Dict[str, List[str]], min: bool = ...) -> None: ...
def ljson(filepath: str) -> Any: ...
def load_jasm(filepath: str) -> Dict[str, List[str]]: ...
def ljasm(filepath: str) -> Dict[str, List[str]]: ...
def lpak(filepath: str) -> Dict[str, List[str]]: ...
def touch(filepath: str) -> None: ...
def shebang(filepath: str) -> Any: ...
